#ifdef REAL
#  define SDOT  ddot
#  define SNRM2 dnrm2
#else
#  define SDOT  sdot
#  define SNRM2 snrm2
#endif

      subroutine cgoptw( n, x, f, g, fcgrd, optline
     $                  , epsx, fmin, epsg, iprint
     $                  , niter, nfunc, ifail
     $                  , d, gold, sbar
     $                  )

      implicit none

c-----------------------------------------
c arguments
c-----------------------------------------
c      n       = dimension
c      x       = startingpoint
c      f       = f(x) value at the starting point
c      x       = starting point
c      g       = valie of the gradient at point x
c      ggbfun  = routine that returns f(x) d[f(x)]
c      optline = linesearch routine
c      epsx    = precision of x
c      fmin    = expected optimal function value  
c      epsg    = precision of g
c      iprint  = switch for information printing
c      niter   = max number of iteration (in)
c                processed number of iteration (out)
c      nfunc   = max number of simulations
c      ifail   = number of failure
c      d       = SEARCH DIRECTION
c      gold    = old d[f(x)]  
c-----------------------------------------

      integer n, iprint, niter, nfunc, ifail, indic
      REAL    x(n), f, g(n), epsx, fmin, epsg

      REAL    d(n), gold(n), sbar(n)

      external fcgrd, optline

c-----------------------------------------
c local variables
c-----------------------------------------
c     j         = counter
c     itmax     = store max iteration == niter 
c     moderl    = error code of the line search
c     ifunc     = processed function evulations
c     r1        = dotdg 
c     tact      = step size
c     tmin      = lower bound of search intervall
c     tmax      = upper bound of search intervall
c     gnorm     = 2-norm of the gradient
c     gnorm0    = 2-norm of the startgradient
c     eps1      = stopping criteria
c     precos    = the result of the fomular from Polak-Rebiere
c     dotdg     = <d,g>
c-----------------------------------------

      integer i, itmax, moderl, ifunc
      REAL   r1, tact, tmin, tmax, gnorm, gnorm0, eps1
      REAL   precos,dotdg 

      external SDOT, SNRM2 
      REAL     SDOT, SNRM2

c-----------------------------------------
c parameters
c-----------------------------------------
      REAL       rm1         , rm2
      parameter( rm1 = 0.0001, rm2=0.9 )
      REAL       rmin
      parameter( rmin = 1.e-20 )
      character*(*) iform
      parameter(iform='(i5,3x,1pe7.1,1x,i5,5x,1pe10.4,3(3x,1pe7.1))' )

c-----------------------------------------
c initialization
c-----------------------------------------

      itmax = niter
      niter = 0
      ifunc = 1
      eps1  = 1.0

c-----------------------------------------
c get first values for f(x) and D[f(x)]
c-----------------------------------------
      call fcgrd( indic, n, x, f, g )

      gnorm0 = SNRM2( n, g )
      if (gnorm0 .lt. rmin) then
         ifail = 3
         goto 1000
      endif

      if (iprint .ge. 1) then
         print '(2a)', '  Itn      Step  Nfun      Objective    '
     $        , 'Norm G    Norm X   Norm (X(k-1)-X(k))'
      end if

c-----------------------------------------
c compute first descent direction
c-----------------------------------------

         precos = 2.*abs(f-fmin)/(gnorm0*gnorm0)
         do i = 1,n
            d(i)    = -g(i)
            sbar(i) = 1.
         end do

c-----------------------------------------
c initialization for the linesearch
c-----------------------------------------
      tmax = 1.0e+10
      tact = 0.1
      dotdg = SDOT( n, d, g )
      if (dotdg .ge. 0.0) then
         ifail = 4
         goto 1000
      end if
      moderl = 0

c-----------------------------------------
c print information line
c-----------------------------------------
      if (iprint .ge. 1) then
         print iform, niter, tact, ifunc, f, gnorm0
     $        , SNRM2( n, x ), tact*SNRM2( n, d )
      end if

c-----------------------------------------
c begin of loop
c compute x(k+1) out of x(k) + t*d,
c where  t > 0.
c-----------------------------------------
 100  continue
      niter = niter+1

c-----------------------------------------
c restart with d = -g
c-----------------------------------------
c      if (MOD( niter,n) .eq. 0) then
c      do i = 1,n
c        d(i) = -g(i)
c      end do
c      end if


c-----------------------------------------
c store old values
c-----------------------------------------
      do i = 1,n
         gold(i) = g(i)
      end do
      
c-----------------------------------------
c compute tmin
c-----------------------------------------
      tmin = 0.
      do i = 1,n
         tmin = max( tmin, abs(d(i)) )
      end do
      tmin = epsx/tmin
      tact  = 0.1
      r1    = dotdg

c-----------------------------------------
c optline returns new values of x, f, and g
c-----------------------------------------
      call optline( n, fcgrd, x, f, r1, tact, tmin, tmax
     $            , d, g, rm2, rm1
     $            , iprint, moderl, ifunc, nfunc, sbar )

      gnorm = SNRM2( n, g )

c-----------------------------------------
c print information line
c-----------------------------------------

      if (iprint .ge. 1) then
         print iform, niter, tact, ifunc, f, gnorm
     $        , SNRM2( n, x ), tact*SNRM2( n, d )
      end if

c-----------------------------------------
c check output mode of mlis0
c-----------------------------------------
      if (moderl.ne.0) then

c-----------------------------------------
c user request abort
c indic < 0 :
c could not compute function at this point
c-----------------------------------------
         if (moderl .lt. 0) then
            ifail = moderl

c-----------------------------------------
c descent is limited by tmax
c-----------------------------------------
         else if (moderl .eq. 1) then
            ifail = 7

c-----------------------------------------
c maximal number of simultation reached
c-----------------------------------------
         else if (moderl .eq. 4) then
            ifail = 6

c-----------------------------------------
c user request abort
c indic = 0
c-----------------------------------------
         else if (moderl .eq. 5) then
            ifail = -1

c-----------------------------------------
c stoped on epsx
c-----------------------------------------
         else if (moderl .eq. 6) then
            ifail = 8
         endif

          goto 1000
      endif

c-----------------------------------------
c test convergence and stopping criteria
c-----------------------------------------
      eps1 = gnorm / gnorm0

      if (eps1 .lt. epsg) then
         ifail = 0
         goto 1000
      endif

      if (niter .eq. itmax) then
         ifail = 5
         goto 1000
      endif

      if (ifunc .gt. nfunc) then
         ifail = 6
         goto 1000
      endif

c-----------------------------------------
c compute new descent direction
c   d = - H.g
c-----------------------------------------
      precos = 0. 
      do i = 1,n
         precos = precos+ (((g(i)-gold(i))*g(i))/(gold(i)*gold(i)))
c        precos = precos+ (g(i)*g(i))/(gold(i)*gold(i))
c        precos = precos+ (g(i)*g(i))/(gold(i)*d(i))
      end do

      do i = 1,n
         d(i) = -g(i) + precos*d(i)
      end do

c-----------------------------------------
c is the new direction a descent one
c if not, force a decent on by -g
c    if this fails, exit with error message
c-----------------------------------------
      dotdg = SDOT( n, d, g )
      if (dotdg .ge. 0.0 ) then
         do i = 1,n
            d(i) = -g(i) 
         end do         
CC      if (iprint .ge. 10) then
CC         print *,'----------------------- force a descent
CC     & direction'
CC      end if
         dotdg = SDOT( n, d, g )
         if (dotdg .ge. 0.0 ) then
            ifail = 4
            goto 1000
         endif
      endif

c-----------------------------------------
c return of loop
c-----------------------------------------
      goto 100

c-----------------------------------------
c loop exit
c-----------------------------------------
 1000 continue
      nfunc = ifunc
      epsg = eps1

      return
      end
