/*******************************************************************/
/*          G E T K E Y                                            */
/*          ===========                                            */
/*                                                                 */
/* purpose       : GETKEY returns a uniq record number to a given  */
/*                 triple (a,b,c) for each (tape,var) pair         */
/*                                                                 */
/* implementaion : binary search trees are used for accessing      */
/*                 (tape,var) pairs and for holding the            */
/*                 triplets (a,b,c)                                */
/*                                                                 */
/* comments      : binary search trees are handled by tsearch      */
/*                 The C-library tsearch is very slow!!            */
/*                 Thus I have replaced it by an implementation    */
/*                 using 'ausgeglichene' binary trees              */
/*                                                                 */
/*                                                                 */
/*******************************************************************/

#include "define.h"
#include "tsearch.h"

#ifdef CRAY
# define GETKEY GETKEY
#else
# ifdef AIX
#  define GETKEY getkey
# else
#  define GETKEY getkey_
# endif
#endif


#define MAXKEY 3

typedef  int    (*Comp) PP((Const void *, Const void *));
typedef  Void   (*Action) PP((Void *, VISIT, int));

struct keynode { 
  FORINT key[MAXKEY];
  FORINT irec;
};


struct keynode keynode_space;
struct keynode *keynodeptr = &keynode_space;


/***********************************************************/
struct tape {
  FORINT k1, k2;
  FORINT maxrec;
  struct keynode *root;
};

struct tape tape_space;
struct tape *tapeptr = &tape_space;
struct tape *tape_root = NULL;


/***********************************************************/
/*          NODE_COMPARE                                   */
/***********************************************************/
Inline
Static int
keynode_compare(node1, node2)
     Const struct keynode *node1, *node2;
{
  int i;

  for( i = MAXKEY-1; i>=0; i--){
    if (node1->key[i] > node2->key[i]) {
      return 1;
    }
    else if (node1->key[i] < node2->key[i]) {
      return -1;
    }
  }
  return 0;
}


/***********************************************************/
/*          PRINT_ENTRY                                    */
/***********************************************************/
Inline
Void
print_entry(node, order, level)
     struct keynode **node;
     VISIT order;
     int level;
{
  if (order == postorder || order == leaf) {
    (Void) printf( "keys = %i, %i, %i  rec = %l\n"
		  , (*node)->key[0], (*node)->key[1], (*node)->key[2]
		  , (*node)->irec);
  }
}



/***********************************************************/
/*          TAPE_COMPARE                                   */
/***********************************************************/
Inline
Static int
tape_compare(node1, node2)
     struct tape *node1, *node2;
{
  if (node1->k1 == node2->k1)
    return (int) (node2->k2 - node1->k2);
  else
    return (int) (node2->k1 - node1->k1);
}

/***********************************************************/
/*          GETTAPE                                        */
/***********************************************************/
Inline
Static struct tape *
gettape( tape1, tape2 )
     FORINT  *tape1, *tape2;
{
  struct tape **found;

  tapeptr->k1 = *tape1;
  tapeptr->k2 = *tape2;
  found = (struct tape **)
    tsearch( (Const Void *) tapeptr, (Void **) (&tape_root)
	    , (Comp) tape_compare );
  if (*found == tapeptr) {

    (*found)->root   = NULL;
    (*found)->maxrec = 0;

    if ((tapeptr = (struct tape *) malloc( sizeof(struct tape) )) == NULL )
      (Void) fprintf( stderr, "gettape : malloc failed\n" );
  }
  return (*found);
}


/***********************************************************/
/*          GETKEY                                         */
/***********************************************************/
Void
GETKEY( tape1, tape2, i1, i2, i3, nrec )
     FORINT  *tape1, *tape2, *i1, *i2, *i3, *nrec;
{
  struct tape *cptr;
  struct keynode **found;

  *nrec = 0;
  cptr = gettape( tape1, tape2 );

  keynodeptr->key[0] = *i1;
  keynodeptr->key[1] = *i2;
  keynodeptr->key[2] = *i3;
 
  found = (struct keynode **)
    tsearch( (Const Void *) keynodeptr, (Void **) (&(cptr->root))
	    , (Comp) keynode_compare );

  if (*found == keynodeptr) {    /* node was inserted, so get a new one */
    keynodeptr->irec = ++cptr->maxrec;
    if ((keynodeptr = (struct keynode *) malloc(sizeof(struct keynode)))
	== NULL )
      (Void) fprintf( stderr, "GETKEY : malloc failed\n" );
  }

  *nrec = (*found)->irec;
/*
  printf("===== tree output =========\n");
  twalk((Anyptr)(cptr->root), (Action) print_entry);
  printf("===========================\n");
*/
}
